/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "surface/surface.pmh"

#define FE_SAJ_DEBUG		FALSE

namespace fe
{

void SurfaceAccessibleJoint::startJoints(void)
{
	SurfaceAccessibleI::Creation creation=SurfaceAccessibleI::e_createMissing;

	m_spOutputPoint=accessor(e_point,e_position,creation);
	m_spOutputColor=accessor(e_point,"Cd",creation);
	m_spOutputNormal=accessor(e_primitive,"N",creation);
	m_spOutputName=accessor(e_primitive,"name",creation);
	m_spOutputParent=accessor(e_primitive,"parent",creation);
	m_spOutputRefX=accessor(e_primitive,"refX",creation);
	m_spOutputRefY=accessor(e_primitive,"refY",creation);
	m_spOutputRefZ=accessor(e_primitive,"refZ",creation);
	m_spOutputRefT=accessor(e_primitive,"refT",creation);
	m_spOutputAnimX=accessor(e_primitive,"animX",creation);
	m_spOutputAnimY=accessor(e_primitive,"animY",creation);
	m_spOutputAnimZ=accessor(e_primitive,"animZ",creation);
	m_spOutputAnimT=accessor(e_primitive,"animT",creation);
	m_spOutputVertices=accessor(e_primitive,e_vertices,creation);

	m_nameList.clear();
	m_parentMap.clear();
	m_refMap.clear();
	m_refTweakMap.clear();
	m_animMap.clear();
	m_animTweakMap.clear();
	m_targetMap.clear();
}

BWORD SurfaceAccessibleJoint::setJoint(
	String a_name,String a_parentName,
	const SpatialTransform& a_rXformRelative,
	const SpatialTransform& a_rRefWorld,Color a_color)
{
#if FE_SAJ_DEBUG
	feLog("SurfaceAccessibleJoint::addJoint\n");
#endif

	SpatialTransform animParent;

	const BWORD hasParent=(m_animMap.find(a_parentName)!=m_animMap.end());
	if(hasParent)
	{
		animParent=m_animMap[a_parentName];
	}
	else
	{
		setIdentity(animParent);
	}

	const SpatialTransform animWorld=a_rXformRelative*animParent;

	m_refMap[a_name]=a_rRefWorld;
	m_animMap[a_name]=animWorld;

	if(m_spFilter.isValid() && m_spFilter->filter(a_name).empty())
	{
		m_refTweakMap[a_name]=a_rRefWorld;
		m_animTweakMap[a_name]=animWorld;
		return FALSE;
	}

	m_nameList.push_back(a_name);
	m_parentMap[a_name]=a_parentName;
	m_colorMap[a_name]=a_color;

	//* NOTE only first one counts
	if(m_targetMap.find(a_parentName)==m_targetMap.end())
	{
		m_targetMap[a_parentName]=a_rRefWorld;
	}
}

void SurfaceAccessibleJoint::completeJoints(void)
{
	const U32 jointCount=m_nameList.size();
	for(U32 jointIndex=0;jointIndex<jointCount;jointIndex++)
	{
		const String& jointName=m_nameList[jointIndex];
		const String& parentName=m_parentMap[jointName];

		const SpatialTransform& refWorld=m_refMap[jointName];
		const SpatialTransform& animWorld=m_animMap[jointName];

		SpatialTransform xformParentRefTweak;
		SpatialTransform xformParentTweak;
		SpatialTransform delta;

		const BWORD hasParent=(m_animMap.find(parentName)!=m_animMap.end());

		if(hasParent)
		{
			xformParentRefTweak=m_refTweakMap[parentName];
			xformParentTweak=m_animTweakMap[parentName];

			SpatialTransform invRefWorld;
			invert(invRefWorld,refWorld);

			delta=invRefWorld*animWorld;
		}
		else
		{
			setIdentity(xformParentRefTweak);
			setIdentity(xformParentTweak);
			setIdentity(delta);
		}

		//* find local tweak in reference

		SpatialTransform invParentRefTweak;
		invert(invParentRefTweak,xformParentRefTweak);

		SpatialTransform refTweak=xformParentRefTweak;

		const SpatialVector toWorld=
				(m_targetMap.find(jointName)==m_targetMap.end())?
				refWorld.translation()-xformParentRefTweak.translation():
				m_targetMap[jointName].translation()-refWorld.translation();

//		feLog("\"%s\" at %s target %s to %s\n",
//				jointName.c_str(),c_print(refWorld.translation()),
//				c_print((m_targetMap.find(jointName)==m_targetMap.end())?
//				SpatialVector(-1,-1,-1): m_targetMap[jointName].translation()),
//				c_print(toWorld));

		const SpatialVector toLocal=rotateVector(invParentRefTweak,toWorld);
		if(magnitudeSquared(toLocal)>1e-6)
		{
			const SpatialVector fromLocal(0.0,1.0,0.0);

			SpatialQuaternion quatLocal;
			set(quatLocal,fromLocal,unitSafe(toLocal));

			const SpatialTransform localTweak=quatLocal;

			refTweak=localTweak*refTweak;
		}
		setTranslation(refTweak,refWorld.translation());

		SpatialTransform animTweak=refTweak*delta;
		setTranslation(animTweak,m_animMap[jointName].translation());

		m_refTweakMap[jointName]=refTweak;
		m_animTweakMap[jointName]=animTweak;

		SpatialTransform invParentTweak;
		invert(invParentTweak,xformParentTweak);

		SpatialTransform xformLocal=animTweak*invParentTweak;
		SpatialTransform xformLocalRef=refTweak*invParentRefTweak;

		addJoint(jointName,parentName,xformLocal,animTweak,
				xformLocalRef,m_colorMap[jointName]);
	}
}

BWORD SurfaceAccessibleJoint::addJoint(String a_name,String a_parentName,
	const SpatialTransform& a_rAnimLocal,
	const SpatialTransform& a_rAnimWorld,
	const SpatialTransform& a_rRefLocal,
	Color a_color)
{
	const BWORD success=addJoint(a_name,a_parentName,
			a_rAnimLocal,a_rAnimWorld,a_color);

	if(success)
	{
		const I32 primitiveIndex=m_spOutputVertices->count()-1;

		m_spOutputRefX->set(primitiveIndex,a_rRefLocal.column(0));
		m_spOutputRefY->set(primitiveIndex,a_rRefLocal.column(1));
		m_spOutputRefZ->set(primitiveIndex,a_rRefLocal.column(2));
		m_spOutputRefT->set(primitiveIndex,a_rRefLocal.translation());
	}

	return success;
}

BWORD SurfaceAccessibleJoint::addJoint(String a_name,String a_parentName,
	const SpatialTransform& a_rAnimLocal,
	const SpatialTransform& a_rAnimWorld,
	Color a_color)
{
	//* TODO param
	const Real triScale=1.0;

	const Color red(1.0,0.0,0.0);
	const Color green(0.0,1.0,0.0);
	const Color blue(0.0,0.0,1.0);

	const I32 primitiveIndex=m_spOutputVertices->append(3);
	m_spOutputName->set(primitiveIndex,a_name);
	m_spOutputParent->set(primitiveIndex,a_parentName);
	m_spOutputNormal->set(primitiveIndex,a_rAnimWorld.column(1));

	const I32 pointIndex0=m_spOutputVertices->integer(primitiveIndex,0);
	const I32 pointIndex1=m_spOutputVertices->integer(primitiveIndex,1);
	const I32 pointIndex2=m_spOutputVertices->integer(primitiveIndex,2);

	const SpatialVector point=a_rAnimWorld.translation();
	m_spOutputPoint->set(pointIndex0,point);
	m_spOutputPoint->set(pointIndex1,point+triScale*a_rAnimWorld.column(0));
	m_spOutputPoint->set(pointIndex2,point+triScale*a_rAnimWorld.column(2));

	m_spOutputColor->set(pointIndex0,a_color);
	m_spOutputColor->set(pointIndex1,red);
	m_spOutputColor->set(pointIndex2,blue);

	m_spOutputAnimX->set(primitiveIndex,a_rAnimLocal.column(0));
	m_spOutputAnimY->set(primitiveIndex,a_rAnimLocal.column(1));
	m_spOutputAnimZ->set(primitiveIndex,a_rAnimLocal.column(2));
	m_spOutputAnimT->set(primitiveIndex,a_rAnimLocal.translation());

	return TRUE;
}

} // namespace
