/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_STA_DEBUG	FALSE
#define	FE_STA_PROFILE	FALSE

namespace fe
{

void SurfaceTrianglesAccessible::bind(
	sp<SurfaceAccessibleI> a_spSurfaceAccessibleI)
{
	m_spSurfaceAccessibleI=a_spSurfaceAccessibleI;
}

void SurfaceTrianglesAccessible::cache(void)
{
	//* populate surface with generic accessors

#if FE_STA_PROFILE
	sp<Profiler> spProfiler(new Profiler("SurfaceTrianglesAccessible"));
	sp<Profiler::Profile> spProfilePrep(
			new Profiler::Profile(spProfiler,"Prep"));
	sp<Profiler::Profile> spProfileLoop(
			new Profiler::Profile(spProfiler,"Loop"));
	sp<Profiler::Profile> spProfileResolve(
			new Profiler::Profile(spProfiler,"Resolve"));
	spProfiler->begin();
	spProfilePrep->start();
#endif

	sp<SurfaceAccessorI> spVertices=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	sp<SurfaceAccessorI> spPointPosition=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);
	sp<SurfaceAccessorI> spPointNormal=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_normal);
	sp<SurfaceAccessorI> spPointUV=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,"uv");
	sp<SurfaceAccessorI> spPrimitivePartition;
	if(!m_partitionAttr.empty())
	{
		feLog("SurfaceTrianglesAccessible::cache partitionAttr \"%s\"\n",
				m_partitionAttr.c_str());

		spPrimitivePartition=m_spSurfaceAccessibleI->accessor(
				SurfaceAccessibleI::e_primitive,m_partitionAttr.c_str());
	}

	const U32 primitiveCount=spVertices->count();

#if FE_STA_DEBUG
		feLog("SurfaceTrianglesAccessible::cache primitiveCount %d\n",
				primitiveCount);
#endif

	m_elements=primitiveCount;
	m_vertices=0;
	U32 elementAllocated=0;
	U32 vertexAllocated=0;

	U32 triangleIndex=0;
	U32 total=0;

	const SpatialVector zero(0.0,0.0,0.0);

	//* pre-allocate estimate
	resizeFor(primitiveCount*2,primitiveCount*6,
			elementAllocated,vertexAllocated);

#if FE_STA_PROFILE
	spProfileLoop->replace(spProfilePrep);
#endif

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spVertices->subCount(primitiveIndex);

#if FE_STA_DEBUG
		feLog("SurfaceTrianglesAccessible::cache %d/%d tri %d/%d verts %d\n",
				primitiveIndex,primitiveCount,total,vertexAllocated,subCount);
#endif

		if(subCount<3)
		{
			continue;
		}

		m_vertices+=(subCount==4)? 6: 3;

		if(m_vertices>vertexAllocated)
		{
			resizeFor(primitiveCount*2,m_vertices,
					elementAllocated,vertexAllocated);
		}

		FEASSERT(total<vertexAllocated);

		I32 partitionIndex= -1;
		if(spPrimitivePartition.isValid())
		{
			const String partition=
					spPrimitivePartition->string(primitiveIndex);
			partitionIndex=lookupPartition(partition);

//			feLog("SurfaceTrianglesAccessible::cache partition %d \"%s\"\n",
//					partitionIndex,partition.c_str());
		}

		m_pElementArray[triangleIndex]=Vector3i(total,3,FALSE);

		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=
					spVertices->integer(primitiveIndex,subIndex);

			m_pVertexArray[total]=
					spPointPosition->spatialVector(pointIndex);
			m_pNormalArray[total]=(spPointNormal.isValid())?
					spPointNormal->spatialVector(pointIndex): zero;
			m_pUVArray[total]=(spPointUV.isValid())?
					spPointUV->spatialVector(pointIndex): zero;

			m_pTriangleIndexArray[total]=triangleIndex+(subIndex==3);
			m_pPointIndexArray[total]=pointIndex;
			m_pPrimitiveIndexArray[total]=primitiveIndex;
			m_pPartitionIndexArray[total]=partitionIndex;

#if FE_STA_DEBUG
			feLog("  %d/%d v %s n %s uv %s tri %d pt %d pr %d part %d\n",
					subIndex,subCount,
					c_print(m_pVertexArray[total]),
					c_print(m_pNormalArray[total]),
					c_print(m_pUVArray[total]),
					m_pTriangleIndexArray[total],
					m_pPointIndexArray[total],
					m_pPrimitiveIndexArray[total],
					m_pPartitionIndexArray[total]);
#endif

			total++;

			if(subIndex==3)
			{
				triangleIndex++;

				m_pElementArray[triangleIndex]=
						Vector3i(total-1,3,FALSE);

				m_pVertexArray[total]=m_pVertexArray[total-4];
				m_pNormalArray[total]=m_pNormalArray[total-4];
				m_pUVArray[total]=m_pUVArray[total-4];
				m_pTriangleIndexArray[total]=triangleIndex;
				m_pPointIndexArray[total]=m_pPointIndexArray[total-4];
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				total++;

				m_pVertexArray[total]=m_pVertexArray[total-3];
				m_pNormalArray[total]=m_pNormalArray[total-3];
				m_pUVArray[total]=m_pUVArray[total-3];
				m_pTriangleIndexArray[total]=triangleIndex;
				m_pPointIndexArray[total]=m_pPointIndexArray[total-3];
				m_pPrimitiveIndexArray[total]=primitiveIndex;
				m_pPartitionIndexArray[total]=partitionIndex;
				total++;
			}
		}

		triangleIndex++;
	}

#if FE_STA_PROFILE
	spProfileResolve->replace(spProfileLoop);
#endif

	m_elements=triangleIndex;
	m_vertices=total;

	if(elementAllocated!=triangleIndex || vertexAllocated!=m_vertices)
	{
		resizeArrays(triangleIndex,m_vertices);
	}

	calcBoundingSphere();

#if FE_STA_PROFILE
	spProfileResolve->finish();
	spProfiler->end();

	feLog("%s:\n%s\n",spProfiler->name().c_str(),
			spProfiler->report().c_str());
#endif
}

} // namespace
